import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import VRP


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool

class TestQUBO(unittest.TestCase):
    """Tests for VRP class"""
    def test_vrp_terms_weights_constant(self):
        """Testing VRP problem creation"""
        pos = [[4, 1], [4, 4], [3, 3]]  # nodes position x, y
        n_vehicles = 1
        n_nodes = len(pos)
        G = nx.Graph()
        G.add_nodes_from(range(n_nodes))
        for i in range(n_nodes - 1):
            for j in range(i + 1, n_nodes):
                r = np.sqrt((pos[i][0] - pos[j][0]) ** 2 + (pos[i][1] - pos[j][1]) ** 2)
                G.add_weighted_edges_from([(i, j, r)])
        vrp_qubo = VRP(G, pos, n_vehicles).qubo
        expected_terms = [[0, 1], [0, 2], [1, 2], [0], [1], [2]]
        expected_weights = [2.0, 2.0, 2.0, 6.5, 6.881966011250105, 7.292893218813452]
        expected_constant = 21.32514076993644

        self.assertTrue(terms_list_equality(expected_terms, vrp_qubo.terms))
        self.assertEqual(expected_weights, vrp_qubo.weights)
        self.assertEqual(expected_constant, vrp_qubo.constant)

    def test_vrp_random_instance(self):
        """Testing the random_instance method of the VRP problem class"""
        seed = 1234
        np.random.seed(seed)
        n_nodes = 3
        n_vehicles = 1
        G = nx.Graph()
        G.add_nodes_from(range(n_nodes))
        pos = [[0, 0]]
        pos += [list(2 * np.random.rand(2) - 1) for _ in range(n_nodes - 1)]
        for i in range(n_nodes - 1):
            for j in range(i + 1, n_nodes):
                r = np.sqrt((pos[i][0] - pos[j][0]) ** 2 + (pos[i][1] - pos[j][1]) ** 2)
                G.add_weighted_edges_from([(i, j, r)])

        vrp_prob = VRP(G, pos, n_vehicles).qubo
        vrp_prob_random = VRP.random_instance(
            n_nodes=n_nodes, n_vehicles=n_vehicles, seed=seed
        ).qubo

        self.assertTrue(terms_list_equality(vrp_prob_random.terms, vrp_prob.terms))
        self.assertEqual(vrp_prob_random.weights, vrp_prob.weights)
        self.assertEqual(vrp_prob_random.constant, vrp_prob.constant)

    def test_vrp_random_instance_unbalanced(self):
        """
        Testing the random_instance method of the VRP problem class using the
        unbalanced penalization method
        """
        seed = 1234
        np.random.seed(seed)
        n_nodes = 8
        n_vehicles = 2
        n_vars = n_nodes * (n_nodes - 1) // 2

        vrp_prob_random = VRP.random_instance(
            n_nodes=n_nodes,
            n_vehicles=n_vehicles,
            seed=seed,
            method="unbalanced",
            penalty=3 * [0.1],
        ).qubo

        self.assertTrue(vrp_prob_random.n == n_vars)
        self.assertTrue(vrp_prob_random.weights[0] == 0.05)
        self.assertTrue(vrp_prob_random.weights[-1] == -0.688413325464513)
        self.assertTrue(vrp_prob_random.terms[0] == [0, 1])
        self.assertTrue(vrp_prob_random.terms[-1] == [27])

    def test_vrp_plot(self):
        """
        Testing the random_instance method of the VRP problem class using the
        unbalanced penalization method
        """
        import matplotlib.pyplot as plt

        seed = 1234
        np.random.seed(seed)
        n_nodes = 8
        n_vehicles = 2

        vrp_prob_random = VRP.random_instance(
            n_nodes=n_nodes, n_vehicles=n_vehicles, seed=seed
        )
        vrp_sol = vrp_prob_random.classical_solution()
        vrp_sol_str = vrp_prob_random.classical_solution(string=True)

        self.assertTrue(isinstance(vrp_prob_random.plot_solution(vrp_sol), plt.Figure))
        self.assertTrue(
            isinstance(vrp_prob_random.plot_solution(vrp_sol_str), plt.Figure)
        )

    def test_vrp_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """
        # Wrong method is called
        with self.assertRaises(ValueError) as e:
            VRP.random_instance(n_nodes=6, n_vehicles=2, method="method")
        self.assertEqual(
            "The method 'method' is not valid.",
            str(e.exception),
        )
        # Penalization terms in unblanced method is not equal to three
        with self.assertRaises(ValueError) as e:
            VRP.random_instance(
                n_nodes=6, n_vehicles=2, method="unbalanced", penalty=[0.1]
            )
        self.assertEqual(
            "The penalty must have 3 parameters [lambda_0, lambda_1, lambda_2]",
            str(e.exception),
        )

        # paths with an unfeasible solution
        with self.assertRaises(ValueError) as e:
            vrp = VRP.random_instance(n_nodes=6, n_vehicles=2, seed=1234)
            sol = vrp.classical_solution()
            sol["x_0_1"] = (
                sol["x_0_1"] + 1
            ) % 2  # Changing one value in the solution to make it an unfeasible solution
            vrp.paths_subtours(sol)
        self.assertEqual(
            "Solution provided does not fulfill all the path conditions.",
            str(e.exception),
        )
        # subtours with an unfeasible direction (broke subtour)
        with self.assertRaises(ValueError) as e:
            vrp = VRP.random_instance(
                n_nodes=10, n_vehicles=2, subtours=[[]], seed=1234
            )
            sol = vrp.classical_solution()
            sol["x_1_2"] = (
                sol["x_1_2"] + 1
            ) % 2  # Changing one value in the solution to make it an unfeasible solution
            vrp.paths_subtours(sol)
        self.assertEqual(
            "The subtours in the solution are broken.",
            str(e.exception),
        )
        # Test unfeasible problem does not return classical solution
        with self.assertRaises(ValueError) as e:
            vrp = VRP.random_instance(n_nodes=3, n_vehicles=3, seed=1234)
            sol = vrp.classical_solution()
            vrp.paths_subtours(sol)
        self.assertEqual(
            "Solution not found: integer infeasible.",
            str(e.exception),
        )
        # Test wrong graph input
        with self.assertRaises(TypeError) as e:
            G = [[0, 1, 2], [1, 2, 3]]
            vrp = VRP(G, pos=[[0, 1], [1, 2]], n_vehicles=2)
        self.assertEqual(
            "Input problem graph must be a networkx Graph.",
            str(e.exception),
        )
        # Test different size between graph nodes and x, y positions in pos
        with self.assertRaises(ValueError) as e:
            G = nx.Graph()
            G.add_nodes_from(range(3))
            vrp = VRP(G, pos=[[0, 1], [1, 2]], n_vehicles=2)
        self.assertEqual(
            "The number of nodes in G is 3 while the x, y coordinates in pos is 2",
            str(e.exception),
        )
        # Test different size between colors and number of vehicles
        with self.assertRaises(ValueError) as e:
            seed = 1234
            vrp = VRP.random_instance(n_nodes=6, n_vehicles=2, seed=seed)
            vrp.plot_solution(vrp.classical_solution(), colors=["tab:blue"])
        self.assertEqual(
            "The length of colors 1 and the number of vehicles 2 do not match",
            str(e.exception),
        )


if __name__ == "__main__":
    unittest.main()


from abc import ABC, abstractmethod
import networkx as nx


class Problem(ABC):
    @staticmethod
    @abstractmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the problem.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments

        Returns
        -------
            A random instance of the problem.
        """
        pass

    def __iter__(self):
        for key, value in self.__dict__.items():
            # remove "_" from the beginning of the key if it exists
            new_key = key[1:] if key.startswith("_") else key
            # convert networkx graphs to dictionaries for serialization
            # (to get back to a graph, use nx.node_link_graph)
            new_value = (
                nx.node_link_data(value) if isinstance(value, nx.Graph) else value
            )
            yield (new_key, new_value)

    @property
    def problem_instance(self):
        """
        Returns a dictionary containing the serialization of the class and
        the problem type name, which will be passed as metadata to the QUBO class.

        Returns
        -------
            A dictionary containing the serialization of the class
            and the problem type name.
        """
        return {**{"problem_type": self.__name__}, **dict(self)}



import unittest
import numpy as np
from openqaoa.problems import PortfolioOptimization


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool

class TestPortfolioOptimization(unittest.TestCase):
    """Tests for PortfoliaOptimization class"""
    def test_portfoliooptimization_terms_weights_constant(self):
        """Test terms,weights,constant of QUBO generated by PortfolioOptimization class"""

        po_terms = [[0, 1], [0, 2], [1, 2], [0], [1], [2]]
        po_weights = [0.505, 0.505, 0.505, 0.535, 0.535, 0.535]
        po_constant = 0.8799999999999999

        mu = [0.1, 0.1, 0.1]
        sigma = [[0.01, 0.01, 0.01], [0.01, 0.01, 0.01], [0.01, 0.01, 0.01]]
        risk_factor = 0.1
        budget = 2
        penalty = 1

        qubo = PortfolioOptimization(mu, sigma, risk_factor, budget, penalty).qubo
        terms, weights = qubo.terms, qubo.weights
        constant = qubo.constant

        self.assertEqual(weights, po_weights)
        self.assertEqual(terms, po_terms)
        self.assertEqual(po_constant, constant)

    def test_portfoliooptimization_random_instance(self):
        """Test random instance method of PortfolioOptimization problem class"""
        seed = 1234
        np.random.seed(seed)
        num_assets = 3
        risk_factor = 0.1
        budget = 2
        penalty = 1
        mu_bounds = [-0.1, 0.1]
        sigma_bounds = [-0.01, 0.01]
        mu = [
            (mu_bounds[1] - mu_bounds[0]) * np.random.rand() + mu_bounds[0]
            for _ in range(num_assets)
        ]
        sigma = [[0 for i in range(num_assets)] for j in range(num_assets)]
        for i in range(num_assets):
            for j in range(num_assets):
                sigma[i][j] = (
                    sigma_bounds[1] - sigma_bounds[0]
                ) * np.random.rand() + sigma_bounds[0]

        qubo = PortfolioOptimization(mu, sigma, risk_factor, budget, penalty).qubo

        qubo_random = PortfolioOptimization.random_instance(
            mu_bounds=mu_bounds,
            sigma_bounds=sigma_bounds,
            risk_factor=risk_factor,
            budget=budget,
            seed=seed,
            penalty=penalty,
            num_assets=num_assets,
        ).qubo
        self.assertEqual(qubo.weights, qubo_random.weights)
        self.assertEqual(qubo.terms, qubo_random.terms)
        self.assertEqual(qubo.constant, qubo_random.constant)

    def test_portfoliooptimization_classical_sol(self):
        """Test the portfolio optimization set random instance method classical solution"""

        seed = 1234
        np.random.seed(seed)
        po_sol = PortfolioOptimization.random_instance(
            num_assets=10, seed=seed
        ).classical_solution()

        sol = {
            "asset_0": 0,
            "asset_1": 0,
            "asset_2": 0,
            "asset_3": 1,
            "asset_4": 1,
            "asset_5": 0,
            "asset_6": 0,
            "asset_7": 1,
            "asset_8": 1,
            "asset_9": 1,
        }

        self.assertEqual(po_sol, sol)

    def test_portfoliooptimization_plot(self):
        """Test portfolio optimization random instance method"""
        import matplotlib.pyplot as plt

        seed = 1234
        portfoliooptimization_random_prob = PortfolioOptimization.random_instance(
            num_assets=12, budget=8, seed=seed
        )
        sol = portfoliooptimization_random_prob.classical_solution()
        figure = portfoliooptimization_random_prob.plot_solution(sol)
        self.assertTrue(isinstance(figure, plt.Figure))
        fig, ax = plt.subplots(figsize=(5, 5))
        self.assertTrue(
            portfoliooptimization_random_prob.plot_solution(sol, ax=ax) == None
        )


if __name__ == "__main__":
    unittest.main()


import numpy as np
from scipy.optimize import OptimizeResult


def CANS(
    fun,
    x0,
    args=(),
    maxfev=None,
    stepsize=0.00001,
    n_shots_cost=None,
    n_shots_min=10,
    n_shots_max=None,
    n_shots_budget=None,
    mu=0.99,
    b=1e-06,
    coeffs=None,
    maxiter=100,
    tol=10 ** (-6),
    jac_w_variance=None,
    callback=None,
    **options
):
    # check that the stepsize is small enough
    lipschitz = np.sum(np.abs(coeffs))
    if not stepsize < 2 / lipschitz:
        raise ValueError(
            "Stepsizec is bigger than 2/Lipschitz: it should be smaller than {0:.3g}".format(
                2 / lipschitz
            )
        )

    # initialize variables for the algorithm loop
    chi = np.zeros(len(x0))
    xi = 0
    n_shots = n_shots_min

    bestx = x0
    besty = np.real(fun(bestx, *args))
    funcalls = 1  # tracks no. of function evals.
    niter = 0

    n_shots_used_total = n_shots_cost
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = besty

    while improved and not stop and niter < maxiter:
        # compute gradient and variance
        gradient, variance, n_shots_used = jac_w_variance(testx, n_shots=n_shots)

        # compute gradient descent step
        testx = testx - stepsize * gradient
        testy = np.real(fun(testx, *args))

        # add the number of shots used to the total (used in the
        # gradient computation and the cost function evaluation)
        n_shots_used_total += n_shots_used + n_shots_cost

        # compute n_shots for next step
        chi = mu * chi + (1 - mu) * gradient
        xi = mu * xi + (1 - mu) * np.sum(variance)
        n_shots = int(
            np.ceil(
                2
                * lipschitz
                * stepsize
                * xi
                / (
                    (2 - lipschitz * stepsize)
                    * (np.linalg.norm(chi) ** 2 + b * mu**niter)
                )
            )
        )

        # clip the number of shots
        n_shots = max(n_shots, n_shots_min)
        n_shots = min(n_shots, n_shots_max) if n_shots_max else n_shots

        if np.abs(besty - testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = np.copy(testx)
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

        # if there is a maximum number of shots and we have reached it, stop
        if n_shots_budget != None and not n_shots_used_total < n_shots_budget:
            stop = True
            break

        niter += 1

    return OptimizeResult(
        fun=besty, x=bestx, nit=niter, nfev=funcalls, success=(niter > 1)
    )



import unittest
from unittest.mock import Mock
import numpy as np
import pytest

from braket.circuits import Circuit

from openqaoa.qaoa_components import (
    PauliOp,
    Hamiltonian,
    QAOADescriptor,
    create_qaoa_variational_params,
    QAOAVariationalStandardParams,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.problems import NumberPartition
from openqaoa_braket.backends import DeviceAWS, QAOAAWSQPUBackend


class TestingQAOABraketQPUBackend(unittest.TestCase):

    """This Object tests the QAOA Braket QPU Backend objects, which is tasked with the
    creation and execution of a QAOA circuit for the selected QPU provider and
    backend.

    These tests require authentication through the AWS CLI.
    """

    @pytest.mark.braket_api
    def test_circuit_angle_assignment_qpu_backend(self):
        """
        A tests that checks if the circuit created by the AWS Backend
        has the appropriate angles assigned before the circuit is executed.
        Checks the circuit created on the AWS Backend.
        """

        nqubits = 3
        p = 2
        weights = [1, 1, 1]
        gammas = [0, 1 / 8 * np.pi]
        betas = [1 / 2 * np.pi, 3 / 8 * np.pi]
        shots = 100

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        aws_backend = QAOAAWSQPUBackend(
            qaoa_descriptor, aws_device, shots, None, None, False, 1.0
        )
        qpu_circuit = aws_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = Circuit()
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[0])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[0])
        main_circuit.rx(1, -2 * betas[0])
        main_circuit.rx(2, -2 * betas[0])
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[1])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[1])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[1])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[1])
        main_circuit.rx(1, -2 * betas[1])
        main_circuit.rx(2, -2 * betas[1])
        main_circuit.probability()

        self.assertEqual(main_circuit, qpu_circuit)

    @pytest.mark.braket_api
    def test_circuit_angle_assignment_qpu_backend_w_hadamard(self):
        """
        Checks for consistent if init_hadamard is set to True.
        """

        nqubits = 3
        p = 2
        weights = [1, 1, 1]
        gammas = [0, 1 / 8 * np.pi]
        betas = [1 / 2 * np.pi, 3 / 8 * np.pi]
        shots = 100

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        aws_backend = QAOAAWSQPUBackend(
            qaoa_descriptor, aws_device, shots, None, None, True, 1.0
        )
        qpu_circuit = aws_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = Circuit()
        main_circuit.h(0)
        main_circuit.h(1)
        main_circuit.h(2)
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[0])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[0])
        main_circuit.rx(1, -2 * betas[0])
        main_circuit.rx(2, -2 * betas[0])
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[1])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[1])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[1])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[1])
        main_circuit.rx(1, -2 * betas[1])
        main_circuit.rx(2, -2 * betas[1])
        main_circuit.probability()

        self.assertEqual(main_circuit, qpu_circuit)

    @pytest.mark.braket_api
    def test_prepend_circuit(self):
        """
        Checks if prepended circuit has been prepended correctly.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 100

        # Prepended Circuit
        prepend_circuit = Circuit()
        prepend_circuit.x(0)
        prepend_circuit.x(1)
        prepend_circuit.x(2)

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        aws_backend = QAOAAWSQPUBackend(
            qaoa_descriptor, aws_device, shots, prepend_circuit, None, True, 1.0
        )
        qpu_circuit = aws_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = Circuit()
        main_circuit.x(0)
        main_circuit.x(1)
        main_circuit.x(2)
        main_circuit.h(0)
        main_circuit.h(1)
        main_circuit.h(2)
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[0])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[0])
        main_circuit.rx(1, -2 * betas[0])
        main_circuit.rx(2, -2 * betas[0])
        main_circuit.probability()

        self.assertEqual(main_circuit, qpu_circuit)

    @pytest.mark.braket_api
    def test_append_circuit(self):
        """
        Checks if appended circuit is appropriately appended to the back of the
        QAOA Circuit.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 100

        # Appended Circuit
        append_circuit = Circuit()
        append_circuit.x(0)
        append_circuit.x(1)
        append_circuit.x(2)

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        aws_backend = QAOAAWSQPUBackend(
            qaoa_descriptor, aws_device, shots, None, append_circuit, True, 1.0
        )
        qpu_circuit = aws_backend.qaoa_circuit(variate_params)

        # Standard Decomposition
        main_circuit = Circuit()
        main_circuit.h(0)
        main_circuit.h(1)
        main_circuit.h(2)
        main_circuit.cnot(0, 1)
        main_circuit.rz(1, 2 * gammas[0])
        main_circuit.cnot(0, 1)
        main_circuit.cnot(1, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(1, 2)
        main_circuit.cnot(0, 2)
        main_circuit.rz(2, 2 * gammas[0])
        main_circuit.cnot(0, 2)
        main_circuit.rx(0, -2 * betas[0])
        main_circuit.rx(1, -2 * betas[0])
        main_circuit.rx(2, -2 * betas[0])
        main_circuit.x(0)
        main_circuit.x(1)
        main_circuit.x(2)
        main_circuit.probability()

        self.assertEqual(main_circuit, qpu_circuit)

    @pytest.mark.braket_api
    def test_prepend_exception(self):
        """
        Test that the error catching for a prepend ciruit larger than the problem
        circuit is invalid
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 100

        # Prepended Circuit
        prepend_circuit = Circuit()
        prepend_circuit.x(0)
        prepend_circuit.x(1)
        prepend_circuit.x(2)
        prepend_circuit.x(3)

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)

        # Try to create the variate params
        _ = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        try:
            # Try to create the AWS backend
            _ = QAOAAWSQPUBackend(
                qaoa_descriptor, aws_device, shots, prepend_circuit, None, True, 1.0
            )
        except Exception as e:
            self.assertEqual(
                str(e), "Cannot attach a bigger circuit to the QAOA routine"
            )

    @pytest.mark.braket_api
    def test_exceptions_in_init(self):
        """
        Testing the Exceptions in the init function of the QAOAAWSQPUBackend
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 100

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)

        # Try instantiating the variate params
        _ = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        # If the user's aws credentials is not correct.
        mock_device = Mock()
        mock_device.configure_mock(
            **{
                "check_connection.return_value": False,
                "provider_connected.return_value": False,
                "qpu_connected.return_value": None,
                "n_qubits": 3,
            }
        )

        try:
            QAOAAWSQPUBackend(
                qaoa_descriptor, mock_device, shots, None, None, True, 1.0
            )
        except Exception as e:
            self.assertEqual(str(e), "Error connecting to AWS.")

        # Wrong arn string name
        aws_device = DeviceAWS(
            "arn:aws:braket:::device/quantum-simulator/amazon/invalid_backend_arn"
        )

        try:
            QAOAAWSQPUBackend(qaoa_descriptor, aws_device, shots, None, None, True, 1.0)
        except Exception as e:
            self.assertEqual(
                str(e),
                "Connection to AWS was made. Error connecting to the specified backend.",
            )

        # No device specified
        aws_device = DeviceAWS("")

        try:
            QAOAAWSQPUBackend(qaoa_descriptor, aws_device, shots, None, None, True, 1.0)
        except Exception as e:
            self.assertEqual(
                str(e), "Connection to AWS was made. A device name was not specified."
            )

        # Correct device arn (Errorless)
        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        QAOAAWSQPUBackend(qaoa_descriptor, aws_device, shots, None, None, True, 1.0)

    @pytest.mark.braket_api
    def test_remote_qubit_overflow(self):
        """
        If the user creates a circuit that is larger than the maximum circuit size
        that is supported by the QPU. An Exception should be raised with the
        appropriate error message alerting the user to the error.
        """

        shots = 100

        set_of_numbers = np.random.randint(1, 10, 100).tolist()
        qubo = NumberPartition(set_of_numbers).qubo

        mixer_hamil = X_mixer_hamiltonian(n_qubits=6)
        qaoa_descriptor = QAOADescriptor(qubo.hamiltonian, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "rand"
        )

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        try:
            braket_backend = QAOAAWSQPUBackend(
                qaoa_descriptor, aws_device, shots, None, None, True, 1.0
            )
            braket_backend.expectation(variate_params)
        except Exception as e:
            self.assertEqual(
                str(e),
                "There are lesser qubits on the device than the number of qubits required for the circuit.",
            )

    @pytest.mark.sim
    def test_remote_integration_qpu_run(self):
        """
        Run a toy example in manual mode to make sure everything works as
        expected for a remote backend
        """

        shots = 100

        set_of_numbers = np.random.randint(1, 10, 10).tolist()
        qubo = NumberPartition(set_of_numbers).qubo

        mixer_hamil = X_mixer_hamiltonian(n_qubits=6)
        qaoa_descriptor = QAOADescriptor(qubo.hamiltonian, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "rand"
        )

        aws_device = DeviceAWS("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

        braket_backend = QAOAAWSQPUBackend(
            qaoa_descriptor, aws_device, shots, None, None, True, 1.0
        )
        braket_backend.expectation(variate_params)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from qiskit import QuantumCircuit

from openqaoa.qaoa_components.ansatz_constructor.gates import (
    RY,
    RX,
    RZ,
    CZ,
    CX,
    RXX,
    RYY,
    RZZ,
    RZX,
    CPHASE,
    RiSWAP,
)
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
from openqaoa_qiskit.backends.gates_qiskit import QiskitGateApplicator


class TestingGate(unittest.TestCase):
    
    def setUp(self):
        
        self.qiskit_gate_applicator = QiskitGateApplicator()

    def test_ibm_gates_1q(self):
        
        # Qiskit Gate Applicator
        gate_applicator = self.qiskit_gate_applicator

        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)

        empty_circuit = QuantumCircuit(1)
        llgate = RY(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.ry(np.pi, 0)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(1)
        llgate = RX(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.rx(np.pi, 0)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(1)
        llgate = RZ(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.rz(np.pi, 0)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

    def test_ibm_gates_2q(self):
        
        # Qiskit Gate Applicator
        gate_applicator = self.qiskit_gate_applicator

        # Two Qubit Gate Tests
        empty_circuit = QuantumCircuit(2)
        llgate = CZ(gate_applicator, 0, 1)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.cz(0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(2)
        llgate = CX(gate_applicator, 0, 1)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.cx(0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

#         empty_circuit = QuantumCircuit(2)
#         llgate = CX(gate_applicator)
#         output_circuit = llgate.apply_ibm_gate([0, 1], empty_circuit)

#         test_circuit = QuantumCircuit(2)
#         test_circuit.ry(np.pi / 2, 1)
#         test_circuit.rx(np.pi, 1)
#         test_circuit.cz(0, 1)
#         test_circuit.ry(np.pi / 2, 1)
#         test_circuit.rx(np.pi, 1)

#         self.assertEqual(
#             test_circuit.to_instruction().definition,
#             output_circuit.to_instruction().definition,
#         )

    def test_ibm_gates_2q_w_gates(self):
        
        # Qiskit Gate Applicator
        gate_applicator = self.qiskit_gate_applicator

        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)

        empty_circuit = QuantumCircuit(2)
        llgate = RXX(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rxx(np.pi, 0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(2)
        llgate = RYY(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.ryy(np.pi, 0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(2)
        llgate = RZZ(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rzz(np.pi, 0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(2)
        llgate = RZX(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rzx(np.pi, 0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )

        empty_circuit = QuantumCircuit(2)
        llgate = CPHASE(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.crz(np.pi, 0, 1)

        self.assertEqual(
            test_circuit.to_instruction().definition,
            output_circuit.to_instruction().definition,
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest
import networkx as nx

from openqaoa import RQAOA
from openqaoa.qaoa_components import Hamiltonian
from openqaoa.algorithms.rqaoa.rqaoa_utils import *
from openqaoa.problems import MaximumCut, QUBO
from openqaoa.backends import create_device

"""
Unittest based testing of current implementation of the RQAOA Algorithm
"""


class TestingRQAOA(unittest.TestCase):
    def test_find_parent(self):
        """
        Test of the find_parent function which backtracks the spin_map dictionary to obtain
        the final constraint used to eliminate a given spin.

        The test consists in finding the correct parents and factors of the constraint
        for a given example.
        """

        # Spin map example without correct final dependencies
        spin_map = dict(
            {
                (0, (1, 0)),
                (1, (-1, 3)),
                (2, (1, 5)),
                (3, (-1, 2)),
                (4, (1, 4)),
                (5, (1, 4)),
            }
        )

        # Solution to the problem
        parents = {0: 0, 1: 4, 2: 4, 3: 4, 4: 4, 5: 4}  # spin:parent_spin
        factors = [1, 1, 1, -1, 1, 1]

        comp_parents = {}
        comp_factors = []

        # For each spin compute parent and connecting factor and store them
        for spin in sorted(spin_map.keys()):
            comp_parent, comp_factor = find_parent(spin_map, spin)
            comp_parents.update({spin: comp_parent})
            comp_factors.append(comp_factor)

        # Test function result
        assert np.allclose(
            list(comp_parents.values()), list(parents.values())
        ), f"Computed parent spins are incorrect"
        assert np.allclose(
            comp_factors, factors
        ), f"Computed constraint factors are incorrect"

    def test_spin_mapping(self):
        """
        Test of the function that generates the spin_map dictionary containing all the elimination
        rules.

        The test consists in generating the correct spin_map and set of max terms and costs in
        its final form (accounting for the dependency that determines the elimination) for a
        given example.
        """

        ## Problem definition

        # Number of qubits
        n_qubits = 20

        # Terms and weights of the graph
        edges = [(i, j) for j in range(n_qubits) for i in range(j)]
        weights = [1 for _ in range(len(edges))]

        # Hyperparameters
        problem = QUBO(n=n_qubits, terms=edges, weights=weights)

        ## Testing

        # Example case of the maximal terms and values extracted from the expectation values
        max_terms_and_stats = dict(
            {
                (4,): 0.99,
                (0, 1): -0.89,
                (0, 9): 0.77,
                (1, 9): -0.73,
                (17,): -0.5,
                (2, 4): 0.32,
                (5, 4): 0.29,
                (16,): -0.8,
                (10,): 0.81,
                (16, 10): 0.4,
                (19, 14): -0.47,
            }
        )

        # Exact solution to the problem
        correct_spin_map = dict(
            {
                (0, (1, 0)),
                (1, (-1, 0)),
                (2, (1, None)),
                (3, (1, 3)),
                (4, (1, None)),
                (5, (1, None)),
                (6, (1, 6)),
                (7, (1, 7)),
                (8, (1, 8)),
                (9, (1, 0)),
                (10, (1, None)),
                (11, (1, 11)),
                (12, (1, 12)),
                (13, (1, 13)),
                (14, (1, 14)),
                (15, (1, 15)),
                (16, (-1, None)),
                (17, (-1, None)),
                (18, (1, 18)),
                (19, (-1, 14)),
            }
        )

        # Compute the spin_map and final constraints from the function
        spin_map = spin_mapping(problem, max_terms_and_stats)

        # Check both outputs contain the same number of keys as the correct solution
        assert len(correct_spin_map) == len(
            spin_map
        ), f"Computed spin_map has incorrect length"

        # Test the spin_map matches the correct solution
        for key in correct_spin_map.keys():
            assert (
                correct_spin_map[key][0] == spin_map[key][0]
            ), f"Computed spin_map contains incorrect factor"
            assert (
                correct_spin_map[key][1] == spin_map[key][1]
            ), f"Computed spin_map contains incorrect parent spin"

    def test_max_terms(self):
        """
        Test of the function that selects spin pairs or singlets are used for elimination,
        given the specific elimination number.

        The test consists in finding the correct pairs and singles to eliminate, according
        to the adaptive scheme, for a given example.
        """

        # Maximum number of singles/pairs to eliminate
        n_elim = 5

        # Set of single spin expectation values
        exp_vals_z = np.array([0.33, -0.21, -0.9, 0.06, -0.78])

        # Correlation matrix
        corr_mat = np.array(
            [
                [0.0, 0.01, -0.64, 0.69, 0.48],
                [0.0, 0.0, -0.99, -0.27, 0.03],
                [0.0, 0.0, 0.0, 1.0, -0.22],
                [0.0, 0.0, 0.0, 0.0, 0.37],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        # Correct solution
        max_tc = dict(
            {(2, 3): 1.0, (1, 2): -0.99, (2,): -0.9, (4,): -0.78, (0, 3): 0.69}
        )

        # Computed solution using the function
        comp_max_tc = max_terms(exp_vals_z, corr_mat, n_elim)

        # Confirm the computed solution has same number of items as correct one
        assert len(max_tc) == len(
            comp_max_tc
        ), f"Computed set of singlets/correlations contains incorrect number of elements"

        # Test the function has obtain the correct singlets/pairs with associated values
        for key in max_tc.keys():
            assert (
                max_tc[key] == comp_max_tc[key]
            ), f"Computed set of singlets/correlations contains incorrect values"

    def test_ada_max_terms(self):
        """
        Test of the function that adaptively selects spin pairs or singlets are used for elimination.

        The test consists in finding the correct pairs and singles to eliminate, according
        to the adaptive scheme, for a given example.
        """

        # Maximum number of singles/pairs allowed to be considered for elimination
        n_max = 3

        # Set of single spin expectation values
        exp_vals_z = np.array([0.33, -0.21, -0.9, 0.06, -0.19])

        # Correlation matrix
        corr_mat = np.array(
            [
                [0.0, 0.01, -0.64, 0.69, 0.48],
                [0.0, 0.0, -0.99, -0.27, 0.03],
                [0.0, 0.0, 0.0, 1.0, -0.22],
                [0.0, 0.0, 0.0, 0.0, 0.37],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        # Correct solution
        max_tc = dict({(2, 3): 1.0, (1, 2): -0.99, (2,): -0.9})

        # Computed solution using the function
        comp_max_tc = ada_max_terms(exp_vals_z, corr_mat, n_max)

        # Confirm the computed solution has same number of items as correct one
        assert len(max_tc) == len(
            comp_max_tc
        ), f"Computed set of singlets/correlations contains incorrect number of elements"

        # Test the function has obtain the correct singlets/pairs with associated values
        for key in max_tc.keys():
            assert (
                max_tc[key] == comp_max_tc[key]
            ), f"Computed set of singlets/correlations contains incorrect values"

    def test_final_solution(self):
        """
        Test the function that reconstructs the final solution by backtracking the
        elimination history and computing the energy of the final states.

        The test consists in reconstructing a set of states for a given elimination history
        amnd computing their energies.
        """

        ## Problem definition

        # Number of qubits
        n_qubits = 10

        # Terms and weights of the graph
        edges = [(i, i + 1) for i in range(n_qubits - 1)] + [(0, n_qubits - 1)] + [(0,)]
        weights = [1 for _ in range(len(edges))]

        # Hamiltonian
        hamiltonian = Hamiltonian.classical_hamiltonian(edges, weights, constant=0)

        ## Testing

        # Trial elimination history and ouput of classical solver
        max_terms_and_stats_list = [
            [
                {"singlet": (1,), "bias": 1.0},
                {"pair": (0, 9), "correlation": -1.0},
            ],
            [
                {"pair": (0, 1), "correlation": 1.0},
                {"pair": (0, 7), "correlation": 1.0},
            ],
            [
                {"pair": (0, 1), "correlation": -1.0},
                {"singlet": (5,), "bias": 1.0},
            ],
            [{"pair": (0, 1), "correlation": 1.0}],
        ]

        classical_states = [[1, 0, 1]]

        # Correct solutions
        states = ["1010101010"]
        energies = [-11]
        correct_full_solution = dict(zip(states, energies))

        # Compute solutions
        full_solution = final_solution(
            max_terms_and_stats_list, classical_states, hamiltonian
        )

        # Test the computed solutions
        assert (
            correct_full_solution == full_solution
        ), f"Solution was not computed correctly"

    def test_problem_from_dict(self):
        """
        Test the function that computes a calssical Hamiltonian from a given graph, accounting for approriate
        labelling of the nodes and edges.

        The test consists in generating the correct QUBO problem for a given graph dictionary.
        """

        # Trial graph
        input_dict = dict(
            {
                (): 10,
                (1,): 1,
                (2,): -1,
                (6,): 4,
                (1, 2): 1,
                (2, 5): 2,
                (10, 14): 3,
                (6, 9): 4,
                (6, 14): 5,
                (5, 6): 6,
            }
        )

        # Correct hamiltonian
        correct_dict = dict(
            {
                (0,): 1,
                (1,): -1,
                (3,): 4,
                (0, 1): 1,
                (1, 2): 2,
                (2, 3): 6,
                (3, 4): 4,
                (3, 6): 5,
                (5, 6): 3,
            }
        )
        hamiltonian = Hamiltonian.classical_hamiltonian(
            list(correct_dict.keys()), list(correct_dict.values()), constant=10
        )
        hamiltonian_dict = {
            term.qubit_indices: coeff
            for term, coeff in zip(hamiltonian.terms, hamiltonian.coeffs)
        }

        # Compute hamiltonian from input graph
        comp_problem = problem_from_dict(input_dict)
        comp_hamiltonian = comp_problem.hamiltonian
        comp_hamiltonian_dict = {
            term.qubit_indices: coeff
            for term, coeff in zip(comp_hamiltonian.terms, comp_hamiltonian.coeffs)
        }

        # Test computed Hamiltonian contains the correct terms
        assert (
            hamiltonian_dict == comp_hamiltonian_dict
        ), f"Terms and coefficients in the computed Hamiltonian are incorrect"

        # Test computed Hamiltonian contains the correct terms
        assert np.allclose(
            hamiltonian.constant, comp_hamiltonian.constant
        ), f"Constant in the computed Hamiltonian is incorrect"

    def test_redefine_problem(self):
        """
        Test the function that computes the new QUBO for a reduced problem, given the original
        QUBO encoding the problem and a set of elimination rules via the spin_map.

        The test consists in computing the new Hamiltonian from a given one and a set of elimination rules.
        """

        ## Problem definition

        # Number of qubits
        n_qubits = 10

        # Edges and weights of the graph
        input_edges = [(i, i + 1) for i in range(n_qubits - 1)] + [(0, n_qubits - 1)]
        input_weights = [1 for _ in range(len(input_edges))]

        # Input problem
        input_problem = QUBO(n_qubits, input_edges, input_weights)

        # Input spin map (elimination rules)
        spin_map = dict(
            {
                (0, (1, 0)),
                (1, (-1, 0)),
                (2, (1, 2)),
                (3, (1, 3)),
                (4, (1, 2)),
                (5, (1, 2)),
                (6, (1, 6)),
                (7, (1, 7)),
                (8, (1, 8)),
                (9, (1, 0)),
            }
        )

        # Compute new problem
        comp_problem, _ = redefine_problem(input_problem, spin_map)

        # Compute the new hamiltonian
        comp_hamiltonian = comp_problem.hamiltonian

        ## Testing (Comparing the new hamiltonian with the correct one)

        # Correct edges, weights and hamiltonian for the reduced problem
        edges = [(0, 1), (1, 2), (1, 3), (3, 4), (4, 5), (0, 5)]
        weights = [-1, 2, 1, 1, 1, 1]

        hamiltonian = Hamiltonian.classical_hamiltonian(edges, weights, constant=0)

        # Test computed Hamiltonian contains the correct terms
        assert (
            hamiltonian.terms == comp_hamiltonian.terms
        ), f"Terms in the computed Hamiltonian are incorrect"

        # Test computed Hamiltonian contains the correct terms
        assert np.allclose(
            hamiltonian.coeffs, comp_hamiltonian.coeffs
        ), f"Coefficients in the computed Hamiltonian are incorrect"

        # Test computed Hamiltonian contains the correct terms
        assert np.allclose(
            hamiltonian.constant, comp_hamiltonian.constant
        ), f"Constant in the computed Hamiltonian is incorrect"

    def test_redefine_problem_isolated_nodes(self):
        """
        Testing the redefine problem function for an edge case in which fixing of one spins leads to an isolated one. In the absense of biases, the only way to capture the isolated node is to keep track of eleiminatiosn and compare the old and new registers to it.

        The test recreates the 5-qubits MaxCut QUBO and the spin map and asserts that the new problem is correct.
        """
        test_qubo = QUBO(
            n=6,
            terms=[[0, 1], [2, 3], [2, 5], [3, 4], [4, 5]],
            weights=[2.0, 1.0, 1.0, 1.0, 1.0],
        )
        spin_map = {0: (1, 0), 1: (-1.0, 0), 2: (1, 2), 3: (1, 3), 4: (1, 4), 5: (1, 5)}
        new_problem, new_spin_map = redefine_problem(test_qubo, spin_map)

        assert new_problem.terms == [[0, 1], [0, 3], [1, 2], [2, 3]]

    def test_solution_for_vanishing_instances(self):
        """
        Test the function that computes the solution to the reduced problem in the rare case the problem vanishes after eliminations.
        The test consists of recreating three specific 4-nodes instances with different connectivity and weights, which can be the result of an RQAOA workflow, and computing their classical energy and ground states according to the function.
        """

        # first example in which 0 and 1 are anticorrelated while the other spins are arbitrary
        hamiltonian = Hamiltonian.classical_hamiltonian(
            [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3]], [2.0, 1.0, 1.0, 1.0, 1.0], 0
        )
        spin_map = {0: (1, None), 1: (-1.0, 0), 2: (1, None), 3: (1, None)}

        cl_energy, cl_ground_states = solution_for_vanishing_instances(
            hamiltonian, spin_map
        )

        assert cl_energy == -2.0
        assert cl_ground_states == ["1011"]

        # second example in which 0 and 2 are correlated while the other spins are arbitrary
        hamiltonian = Hamiltonian.classical_hamiltonian(
            [[0, 1], [0, 3], [2, 3], [1, 2], [0, 2]], [1.0, -1.0, 1.0, -1.0, -2.0], 0
        )
        spin_map = {0: (1, None), 1: (1, None), 2: (1.0, 0), 3: (1, None)}

        cl_energy, cl_ground_states = solution_for_vanishing_instances(
            hamiltonian, spin_map
        )

        assert cl_energy == -2.0
        assert cl_ground_states == ["1111"]

        # third example in which 0 and 3 are anticorrelated while the other spins are arbitrary
        hamiltonian = Hamiltonian.classical_hamiltonian(
            [[0, 1], [0, 3], [0, 2], [1, 3], [2, 3]], [1.0, 2.0, -1.0, 1.0, -1.0], 0
        )
        spin_map = {0: (1, None), 1: (1, None), 2: (1, None), 3: (-1.0, 0)}

        cl_energy, cl_ground_states = solution_for_vanishing_instances(
            hamiltonian, spin_map
        )

        assert cl_energy == -2.0
        assert cl_ground_states == ["1110"]

    def workflow_mockup(self, qubo, n_cutoff):
        r = RQAOA()
        r.set_rqaoa_parameters(n_cutoff=n_cutoff)
        r.set_circuit_properties(
            p=1, param_type="standard", init_type="ramp", mixer_hamiltonian="x"
        )

        # Define the device you want to run your problem on using the create_device() function
        device = create_device(location="local", name="analytical_simulator")
        r.set_device(device)

        # Set the classical method used to optimize over QAOA angles and its properties
        r.set_classical_optimizer(method="cobyla", maxiter=600)

        r.compile(qubo)
        r.optimize()
        return r.result

    def test_total_elimination_cutoff_1_whole_workflow(self):
        """
        Testing the whole RQAOA workflow for Solving the MaxCut on the Ring of Disagree with normalized weights and no biases.

        Recreating the specific qubo instance and running standard RQAOA with cutoff set to 1. This results in vanishing problem before reaching the cutoff, which tests both the proper isolation of nodes and the solution_for_vanishing_instances functions. RQAOA should be able to find the true ground state for this specific instance.
        """
        qubo = QUBO(
            n=8,
            terms=[[0, 7], [0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7]],
            weights=[1, 1 / 9, 2 / 9, 3 / 9, 5 / 9, 6 / 9, 7 / 9, 8 / 9],
        )

        opt_results = self.workflow_mockup(qubo, n_cutoff=1)

        assert opt_results["solution"] == {"10101010": -4.555555555555555}

    def test_isolated_nodes_whole_workflow(self):
        """
        Testing the whole RQAOA workflow for solving the MaxCut problem on random graph instances.

        Recreating a specific qubo and testing it with cutoff size set to 3, 2, and 1. RQAOA should be able to find the true ground state for this specific instance at all cutoffs. Note that the smaller the cutoff, the less of the degenerate solutions are recovered.
        """
        terms = [
            [0, 2],
            [0, 3],
            [0, 4],
            [0, 5],
            [0, 6],
            [0, 7],
            [0, 8],
            [0, 9],
            [0, 1],
            [1, 2],
            [1, 5],
            [1, 6],
            [2, 3],
            [2, 5],
            [2, 6],
            [2, 7],
            [8, 2],
            [9, 2],
            [3, 4],
            [3, 5],
            [3, 6],
            [3, 7],
            [8, 3],
            [9, 3],
            [1, 3],
            [4, 5],
            [4, 6],
            [4, 7],
            [8, 4],
            [9, 4],
            [1, 4],
            [5, 6],
            [5, 7],
            [9, 5],
            [6, 7],
            [8, 6],
            [9, 6],
            [8, 7],
            [1, 7],
            [8, 9],
            [8, 1],
            [1, 9],
        ]
        weights = [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            2.0,
            1.0,
            2.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
        ]
        qubo = QUBO(n=10, terms=terms, weights=weights)

        opt_results_3 = self.workflow_mockup(qubo, n_cutoff=3)
        assert opt_results_3["solution"] == {"1011101000": -10.0, "0010101101": -10.0}
        assert opt_results_3["classical_output"]["optimal_states"] == ["110", "001"]

        opt_results_2 = self.workflow_mockup(qubo, n_cutoff=2)
        assert opt_results_2["solution"] == {"0010101101": -10.0, "1011101000": -10.0}
        assert opt_results_2["classical_output"]["optimal_states"] == ["00", "11"]

        opt_results_1 = self.workflow_mockup(qubo, n_cutoff=1)
        assert opt_results_1["solution"] == {"1011101000": -10.0}
        assert opt_results_1["classical_output"]["optimal_states"] == ["11"]


if __name__ == "__main__":
    unittest.main()



from typing import Callable

from qiskit.circuit.library import (
    XGate,
    RXGate,
    RYGate,
    RZGate,
    CXGate,
    CZGate,
    RXXGate,
    RZXGate,
    RZZGate,
    RYYGate,
    CRZGate,
)
from qiskit import QuantumCircuit, QuantumRegister

from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
import openqaoa.qaoa_components.ansatz_constructor.gates as gates_core


class QiskitGateApplicator(gates_core.GateApplicator):
    QISKIT_OQ_GATE_MAPPER = {
        gates_core.X.__name__: XGate,
        gates_core.RZ.__name__: RZGate,
        gates_core.RX.__name__: RXGate,
        gates_core.RY.__name__: RYGate,
        gates_core.CX.__name__: CXGate,
        gates_core.CZ.__name__: CZGate,
        gates_core.RXX.__name__: RXXGate,
        gates_core.RZX.__name__: RZXGate,
        gates_core.RZZ.__name__: RZZGate,
        gates_core.RYY.__name__: RYYGate,
        gates_core.CPHASE.__name__: CRZGate,
    }

    library = "qiskit"

    def create_quantum_circuit(self, n_qubits) -> QuantumCircuit:
        """
        Function which creates and empty circuit specific to the qiskit backend.
        Needed for SPAM twirling but more general than this.
        """
        qureg = QuantumRegister(n_qubits)
        parametric_circuit = QuantumCircuit(qureg)
        return parametric_circuit

    def gate_selector(self, gate: gates_core.Gate) -> Callable:
        selected_qiskit_gate = QiskitGateApplicator.QISKIT_OQ_GATE_MAPPER[gate.__name__]
        return selected_qiskit_gate

    @staticmethod
    def apply_1q_rotation_gate(
        qiskit_gate,
        qubit_1: int,
        rotation_object: RotationAngle,
        circuit: QuantumCircuit,
    ) -> QuantumCircuit:
        circuit.append(qiskit_gate(rotation_object.rotation_angle), [qubit_1], [])
        return circuit

    @staticmethod
    def apply_2q_rotation_gate(
        qiskit_gate,
        qubit_1: int,
        qubit_2: int,
        rotation_object: RotationAngle,
        circuit: QuantumCircuit,
    ) -> QuantumCircuit:
        circuit.append(
            qiskit_gate(rotation_object.rotation_angle), [qubit_1, qubit_2], []
        )
        return circuit

    @staticmethod
    def apply_1q_fixed_gate(
        qiskit_gate, qubit_1: int, circuit: QuantumCircuit
    ) -> QuantumCircuit:
        circuit.append(qiskit_gate(), [qubit_1], [])
        return circuit

    @staticmethod
    def apply_2q_fixed_gate(
        qiskit_gate, qubit_1: int, qubit_2: int, circuit: QuantumCircuit
    ) -> QuantumCircuit:
        circuit.append(qiskit_gate(), [qubit_1, qubit_2], [])
        return circuit

    def apply_gate(self, gate: gates_core.Gate, *args):
        selected_qiskit_gate = self.gate_selector(gate)
        if gate.n_qubits == 1:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,rotation_object,circuit)
                return self.apply_1q_rotation_gate(selected_qiskit_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,circuit)
                return self.apply_1q_fixed_gate(selected_qiskit_gate, *args)
        elif gate.n_qubits == 2:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,qubit_2,rotation_object,circuit)
                return self.apply_2q_rotation_gate(selected_qiskit_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,qubit_2,circuit)
                return self.apply_2q_fixed_gate(selected_qiskit_gate, *args)
        else:
            raise ValueError("Only 1 and 2-qubit gates are supported.")



import unittest
import json
import itertools
import subprocess
import pytest

import openqaoa
from openqaoa_azure.backends import DeviceAzure


class TestingDeviceAzure(unittest.TestCase):

    """These tests check the Object used to access Azure and their
    available QPUs can be established.

    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with Azure.
    """

    @pytest.mark.api
    def setUp(self):

        bashCommand = "az resource list"
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()

        if error is not None:
            print(error)
            raise Exception(
                "You must have the Azure CLI installed and must be logged in to use the Azure Quantum Backends"
            )
        else:
            output_json = json.loads(output)
            output_json_s = [
                each_json
                for each_json in output_json
                if each_json["name"] == "TestingOpenQAOA"
            ][0]
            self.RESOURCE_ID = output_json_s["id"]
            self.AZ_LOCATION = output_json_s["location"]

    @pytest.mark.api
    def test_check_connection_provider_no_resource_id_or_az_location(self):

        """
        If no information about about the workspace is provided, the resource id
        or az location, check_connection and provider_connected should return False.
        """

        for resource_id, az_location in itertools.product(
            ["", self.RESOURCE_ID], ["", self.AZ_LOCATION]
        ):

            if not (
                resource_id == self.RESOURCE_ID and az_location == self.AZ_LOCATION
            ):

                device_obj = DeviceAzure(
                    device_name="", resource_id=resource_id, az_location=az_location
                )

                self.assertEqual(device_obj.check_connection(), False)
                self.assertEqual(device_obj.provider_connected, False)
                self.assertEqual(device_obj.qpu_connected, None)
        
    @pytest.mark.api      
    def test_check_connection_provider_no_backend_provided_resource_id_and_az_location(self):
        
        """
        If no information about the device name, but the resource id and azure 
        location used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceAzure(
            device_name="", resource_id=self.RESOURCE_ID, az_location=self.AZ_LOCATION
        )

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)

    @pytest.mark.api
    def test_check_connection_provider_right_backend_provided_resource_id_and_az_location(self):

        """
        If the correct device name is provided and the resource id and azure 
        location used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceAzure(
            device_name="", resource_id=self.RESOURCE_ID, az_location=self.AZ_LOCATION
        )

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceAzure(
            device_name=valid_qpu_name,
            resource_id=self.RESOURCE_ID,
            az_location=self.AZ_LOCATION,
        )

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)

    @pytest.mark.api
    def test_check_connection_provider_wrong_backend_provided_resource_id_and_az_location(self):
        
        """
        If device name provided is incorrect, and not empty, and the resource id 
        and azure location used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceAzure(
            device_name="invalid_backend",
            resource_id=self.RESOURCE_ID,
            az_location=self.AZ_LOCATION,
        )

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)


if __name__ == "__main__":
    unittest.main()



from .backends import (
    DeviceQiskit,
    QAOAQiskitQPUBackend,
    QAOAQiskitBackendShotBasedSimulator,
    QAOAQiskitBackendStatevecSimulator,
)

device_access = {DeviceQiskit: QAOAQiskitQPUBackend}
device_name_to_obj = {
    "qiskit.qasm_simulator": QAOAQiskitBackendShotBasedSimulator,
    "qiskit.shot_simulator": QAOAQiskitBackendShotBasedSimulator,
    "qiskit.statevector_simulator": QAOAQiskitBackendStatevecSimulator,
}
device_location = "ibmq"
device_plugin = DeviceQiskit
device_args = {DeviceQiskit: ["hub", "group", "project", "as_emulator"]}
backend_args = {
    QAOAQiskitBackendStatevecSimulator: [],
    QAOAQiskitBackendShotBasedSimulator: [
        "n_shots",
        "seed_simulator",
        "qiskit_simulation_method",
        "noise_model",
        "initial_qubit_mapping",
    ],
    QAOAQiskitQPUBackend: [
        "n_shots",
        "initial_qubit_mapping",
        "qiskit_optimization_level",
    ],
}



import unittest

from openqaoa.backends.qaoa_backend import (
    get_qaoa_backend,
    DEVICE_NAME_TO_OBJECT_MAPPER,
)
from openqaoa.qaoa_components import (
    Hamiltonian,
    create_qaoa_variational_params,
    QAOADescriptor,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_device import create_device
from openqaoa.backends.basebackend import QAOABaseBackendShotBased


def get_params():
    cost_hamil = Hamiltonian.classical_hamiltonian([[0, 1]], [1], constant=0)
    mixer_hamil = X_mixer_hamiltonian(2)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variational_params_std = create_qaoa_variational_params(
        qaoa_descriptor, "standard", "ramp"
    )

    return qaoa_descriptor, variational_params_std


class TestingBackendLocal(unittest.TestCase):
    """
    These tests check that the methods of the local backends are working properly.
    """

    def test_get_counts_and_expectation_n_shots(self):
        """
        Check that the .get_counts admit n_shots as an argument, and works properly for the backend of all local devices.
        Also check that .expectation and .expecation_w_uncertainty methods admit n_shots as an argument for the QAOABaseBackendShotBased backends.
        """
        print(DEVICE_NAME_TO_OBJECT_MAPPER)
        for device_name in DEVICE_NAME_TO_OBJECT_MAPPER.keys():

            # Analytical device doesn't have any of those so we are skipping it in the tests.
            if device_name in ["analytical_simulator"]:
                continue

            qaoa_descriptor, variational_params_std = get_params()

            device = create_device(location="local", name=device_name)
            backend = get_qaoa_backend(
                qaoa_descriptor=qaoa_descriptor, device=device, n_shots=1000
            )

            assert (
                sum(
                    backend.get_counts(
                        params=variational_params_std, n_shots=58
                    ).values()
                )
                == 58
            ), "`n_shots` is not being respected for the local simulator `{}` when calling backend.get_counts(n_shots=58).".format(
                device_name
            )
            if isinstance(backend, QAOABaseBackendShotBased):
                try:
                    backend.expectation(params=variational_params_std, n_shots=58)
                except Exception:
                    raise Exception(
                        "backend.expectation does not admit `n_shots` as an argument for the local simulator `{}`.".format(
                            device_name
                        )
                    )
                try:
                    backend.expectation_w_uncertainty(
                        params=variational_params_std, n_shots=58
                    )
                except Exception:
                    raise Exception(
                        "backend.expectation_w_uncertainty does not admit `n_shots` as an argument for the local simulator `{}`.".format(
                            device_name
                        )
                    )


if __name__ == "__main__":
    unittest.main()



# Cost function to be used for QAOA training
from typing import Dict
import numpy as np

from ..qaoa_components import Hamiltonian
from ..utilities import bitstring_energy


def expectation_value_classical(counts: Dict, hamiltonian: Hamiltonian):
    """
    Evaluate the cost function, i.e. expectation value ``$$\langle|H \rangle$$``
    w.r.t to the measurements results ``counts``.

    Parameters
    ----------
    counts: dict
                The counts of the measurements.
    hamiltonian: Hamiltonian
                The Cost Hamiltonian defined for the optimization problem
    """
    shots = sum(counts.values())
    cost = 0
    for basis_state, count in counts.items():
        cost += count * (bitstring_energy(hamiltonian, basis_state) / shots)
    return cost


def cvar_expectation_value_classical(
    counts: Dict, hamiltonian: Hamiltonian, alpha: float
):
    """
    CVaR computation of cost function. For the definition of the cost function, refer
    to https://arxiv.org/abs/1907.04769.

    Parameters
    ----------
    counts: `dict`
            The counts of the measurements.
    hamiltonian: `Hamiltonian`
            The Cost Hamiltonian defined for the optimization problem
    alpha: `float`
            The CVaR parameter.
    """
    assert alpha > 0 and alpha < 1, "Please specify a valid alpha value between 0 and 1"
    shots = sum(counts.values())

    cost_list = []

    # eigen-energy computation of each basis state in counts
    for basis_state, count in counts.items():
        cost_list.append(count * (bitstring_energy(hamiltonian, basis_state) / shots))

    # sort costs in ascending order
    sorted_cost_list = np.sort(np.array(cost_list))

    K = max([int(len(cost_list) * alpha), 1])
    cost = sum(sorted_cost_list[:K])

    return cost


def cost_function(counts: Dict, hamiltonian: Hamiltonian, alpha: float = 1):
    """
    The cost function to be used for QAOA training.

    Parameters
    ----------
    counts: `dict`
            The counts of the measurements.
    hamiltonian: `Hamiltonian`
            The Cost Hamiltonian defined for the optimization problem
    alpha: `float`
            The CVaR parameter.
    """
    if alpha == 1:
        return expectation_value_classical(counts, hamiltonian)
    else:
        return cvar_expectation_value_classical(counts, hamiltonian, alpha)



from .backends import DeviceAWS, QAOAAWSQPUBackend

device_access = {DeviceAWS: QAOAAWSQPUBackend}
device_location = "aws"
device_plugin = DeviceAWS
device_args = {DeviceAWS: ["s3_bucket_name", "aws_region", "folder_name"]}
backend_args = {
    QAOAAWSQPUBackend: ["n_shots", "disable_qubit_rewiring", "initial_qubit_mapping"]
}



import os
from copy import deepcopy
from typing import Optional, List
import warnings

from braket.circuits import Circuit
from braket.circuits.gates import H
from braket.circuits.result_types import Probability
from braket.circuits.free_parameter import FreeParameter
from braket.jobs.metrics import log_metric

from .devices import DeviceAWS
from .gates_braket import BraketGateApplicator
from openqaoa.backends.basebackend import (
    QAOABaseBackendShotBased,
    QAOABaseBackendCloud,
    QAOABaseBackendParametric,
)
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)


class QAOAAWSQPUBackend(
    QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased
):
    """
    A QAOA simulator as well as for real QPU using Amazon Braket as backend

    Parameters
    ----------
    device: `DeviceAWS`
        An object of the class ``DeviceAWS`` which contains the credentials
        for accessing the QPU via cloud and the name of the device.
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.
    append_state: `QuantumCircuit`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.
    cvar_alpha: `float`
        The value of alpha for the CVaR method.
    disable_qubit_rewiring: `bool`
        A boolean that determines whether qubit routing on the provider's end is
        used. This is False by default. Not all providers provide this feature.
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        device: DeviceAWS,
        n_shots: int,
        prepend_state: Optional[Circuit],
        append_state: Optional[Circuit],
        init_hadamard: bool,
        cvar_alpha: float,
        disable_qubit_rewiring: bool = False,
        initial_qubit_mapping: Optional[List[int]] = None,
    ):
        QAOABaseBackendShotBased.__init__(
            self,
            qaoa_descriptor,
            n_shots,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )
        QAOABaseBackendCloud.__init__(self, device)
        self.gate_applicator = BraketGateApplicator()

        self.qureg = list(range(self.n_qubits))
        self.problem_reg = self.qureg[0 : self.problem_qubits]
        if self.initial_qubit_mapping is None:
            self.initial_qubit_mapping = (
                initial_qubit_mapping
                if initial_qubit_mapping is not None
                else list(range(self.n_qubits))
            )
        else:
            warnings.warn(
                "Ignoring the initial_qubit_mapping since the routing algorithm chose one"
            )
        self.disable_qubit_rewiring = disable_qubit_rewiring

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), (
                "Cannot attach a bigger circuit" " to the QAOA routine"
            )

        if self.device.provider_connected and self.device.qpu_connected:
            self.backend_qpu = self.device.backend_device
        elif (
            self.device.provider_connected is True
            and self.device.qpu_connected is False
        ):
            raise Exception(
                "Connection to AWS was made. Error connecting to the specified backend."
            )
        elif (
            self.device.provider_connected is True and self.device.qpu_connected is None
        ):
            raise Exception(
                "Connection to AWS was made. A device name was not specified."
            )
        else:
            raise Exception("Error connecting to AWS.")

        if self.device.n_qubits < self.n_qubits:
            raise Exception(
                "There are lesser qubits on the device than the number of qubits required for the circuit."
            )

        self.parametric_circuit = self.parametric_qaoa_circuit

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Circuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `Circuit`
            The final QAOA circuit constructed using the angles from variational params.
        """
        parametric_circuit = deepcopy(self.parametric_circuit)

        if self.append_state:
            parametric_circuit += self.append_state

        # TODO: needs to be fixed --> measurement operations on problem qubits
        parametric_circuit += Probability.probability()

        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(
            zip(
                [
                    each_free_param_obj.name
                    for each_free_param_obj in self.braket_parameter_list
                ],
                angles_list,
            )
        )
        circuit_with_angles = parametric_circuit.make_bound_circuit(memory_map)

        return circuit_with_angles

    @property
    def parametric_qaoa_circuit(self) -> Circuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.
        """
        parametric_circuit = Circuit()

        if self.prepend_state:
            parametric_circuit += self.prepend_state

        # Initial state is all |+>
        if self.init_hadamard:
            for each_qubit in self.problem_reg:
                parametric_circuit += H.h(each_qubit)

        self.braket_parameter_list = []
        for each_gate in self.abstract_circuit:
            # if gate is of type mixer or cost gate, assign parameter to it
            if each_gate.gate_label.type.value in ["MIXER", "COST"]:
                angle_param = FreeParameter(each_gate.gate_label.__repr__())
                self.braket_parameter_list.append(angle_param)
                each_gate.angle_value = angle_param
            decomposition = each_gate.decomposition("standard")
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0](self.gate_applicator, *each_tuple[1])
                gate.apply_gate(parametric_circuit)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Execute the circuit and obtain the counts

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: int
            The number of times to run the circuit. If None, n_shots is set to the default: self.n_shots.

        Returns
        -------
            A dictionary with the bitstring as the key and the number of counts
            as its value.
        """

        n_shots = self.n_shots if n_shots == None else n_shots

        circuit = self.qaoa_circuit(params)

        job_state = False
        no_of_job_retries = 0
        max_job_retries = 5

        while job_state == False:
            job = self.backend_qpu.run(
                circuit,
                (self.device.s3_bucket_name, self.device.folder_name),
                shots=n_shots,
                disable_qubit_rewiring=self.disable_qubit_rewiring,
            )

            try:
                self.job_id = job.id

                job_result = job.result()

                # If there was an issue with the job sent, send again.
                if job.state() in ["FAILED", "CANCELLED"] or job_result == None:
                    raise ValueError

                counts = job_result.measurement_counts

            except ValueError:
                print("The task has failed or was cancelled by AWS. Resending task.")
                no_of_job_retries += 1

            except Exception as e:
                print(e, "\n")
                print(
                    "An unknown error occurred while trying to retrieve task results. Resending task."
                )
                no_of_job_retries += 1

            else:
                job_state = True

            finally:
                if no_of_job_retries >= max_job_retries:
                    raise ConnectionError(
                        "An Error Occurred with the Task(s) sent to AWS."
                    )

        final_counts = counts
        # if self.final_mapping is not None:
        #     final_counts = permute_counts_dictionary(final_counts,
        #                                             self.final_mapping)
        # # Expose counts
        self.measurement_outcomes = final_counts
        return final_counts

    def log_with_backend(self, metric_name: str, value, iteration_number) -> None:
        """
        If using AWS Jobs, these values will be logged.
        """

        try:
            if os.environ["AMZN_BRAKET_JOB_NAME"] is not None:
                in_jobs = True
        except KeyError:
            in_jobs = False

        if in_jobs:
            log_metric(
                metric_name=metric_name,
                value=value,
                iteration_number=iteration_number,
            )

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the entire QAOA `QuantumCircuit` object into
        a OpenQASM string
        """
        raise NotImplementedError()
        # qasm_string = self.qaoa_circuit(params).qasm(formatted=True)
        # return qasm_string

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation
        """
        raise NotImplementedError()



from typing import List, Optional, Union
import numpy as np
import json
from scipy.optimize._minimize import MINIMIZE_METHODS

from ..optimizers.training_vqa import CustomScipyGradientOptimizer, PennyLaneOptimizer
from ..backends.devices_core import SUPPORTED_LOCAL_SIMULATORS
from ..backends.basebackend import QuantumCircuitBase
from ..utilities import convert2serialize


ALLOWED_PARAM_TYPES = [
    "standard",
    "standard_w_bias",
    "extended",
    "fourier",
    "fourier_extended",
    "fourier_w_bias",
    "annealing",
]
ALLOWED_INIT_TYPES = ["rand", "ramp", "custom"]
ALLOWED_MIXERS = ["x", "xy"]

ALLOWED_MINIMIZATION_METHODS = (
    MINIMIZE_METHODS
    + CustomScipyGradientOptimizer.CUSTOM_GRADIENT_OPTIMIZERS
    + PennyLaneOptimizer.PENNYLANE_OPTIMIZERS
)

ALLOWED_QVM_DEVICES = ["Aspen-11", "Aspen-M-1"]
ALLOWED_QVM_DEVICES.extend(f"{n}q-qvm" for n in range(2, 80))

ALLOWED_LOCAL_SIMUALTORS = SUPPORTED_LOCAL_SIMULATORS
ALLOWED_IMBQ_GLOBAL = [
    "ibmq_qasm_simulator",
    "ibmq_armonk",
    "ibmq_santiago",
    "ibmq_bogota",
    "ibmq_lima",
    "ibmq_belem",
    "ibmq_quito",
    "simulator_statevector",
    "simulator_mps",
    "simulator_extended_stabilizer",
    "simulator_stabilizer",
    "ibmq_manila",
]
ALLOWED_DEVICES = ALLOWED_LOCAL_SIMUALTORS + ALLOWED_QVM_DEVICES + ALLOWED_IMBQ_GLOBAL


class WorkflowProperties:
    def __iter__(self):
        for key, value in self.__dict__.items():
            # remove "_" from the beginning of the key if it exists
            yield (key[1:] if key.startswith("_") else key, value)

    def asdict(self):
        return convert2serialize(dict(self))


class CircuitProperties(WorkflowProperties):
    """
    Tunable properties of the QAOA circuit to be specified by the user
    """

    def __init__(
        self,
        param_type: str = "standard",
        init_type: str = "ramp",
        qubit_register: List = [],
        p: int = 1,
        q: Optional[int] = 1,
        annealing_time: Optional[float] = None,
        linear_ramp_time: Optional[float] = None,
        variational_params_dict: Optional[dict] = {},
        mixer_hamiltonian: Optional[str] = "x",
        mixer_qubit_connectivity: Optional[Union[List[list], List[tuple], str]] = None,
        mixer_coeffs: Optional[float] = None,
        seed: Optional[int] = None,
    ):
        self.param_type = param_type
        self.init_type = init_type
        self.qubit_register = qubit_register
        self.p = p
        self.q = (
            q
            if param_type.lower() in ["fourier", "fourier_extended", "fourier_w_bias"]
            else None
        )
        self.variational_params_dict = variational_params_dict
        self.annealing_time = (
            annealing_time if annealing_time is not None else 0.7 * self.p
        )
        self.linear_ramp_time = (
            linear_ramp_time if linear_ramp_time is not None else 0.7 * self.p
        )
        self.mixer_hamiltonian = mixer_hamiltonian
        if self.mixer_hamiltonian.lower() == "xy":
            self.mixer_qubit_connectivity = (
                mixer_qubit_connectivity
                if mixer_qubit_connectivity is not None
                else "full"
            )
        else:
            self.mixer_qubit_connectivity = None
        self.mixer_coeffs = mixer_coeffs
        self.seed = seed

    @property
    def param_type(self):
        return self._param_type

    @param_type.setter
    def param_type(self, value):
        if value not in ALLOWED_PARAM_TYPES:
            raise ValueError(
                f"param_type {value} is not recognised. Please use {ALLOWED_PARAM_TYPES}"
            )
        self._param_type = value

    @property
    def init_type(self):
        return self._init_type

    @init_type.setter
    def init_type(self, value):
        if value not in ALLOWED_INIT_TYPES:
            raise ValueError(
                f"init_type {value} is not recognised. Please use {ALLOWED_INIT_TYPES}"
            )
        self._init_type = value

    @property
    def mixer_hamiltonian(self):
        return self._mixer_hamiltonian

    @mixer_hamiltonian.setter
    def mixer_hamiltonian(self, value):
        if value not in ALLOWED_MIXERS:
            raise ValueError(
                f"mixer_hamiltonian {value} is not recognised. Please use {ALLOWED_MIXERS}"
            )
        self._mixer_hamiltonian = value

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, value):
        if value <= 0:
            raise ValueError(
                f"Number of layers `p` cannot be smaller or equal to zero. Value {value} was provided"
            )
        self._p = value

    @property
    def annealing_time(self):
        return self._annealing_time

    @annealing_time.setter
    def annealing_time(self, value):
        if value <= 0:
            raise ValueError(
                f"The annealing time `annealing_time` cannot be smaller or equal to zero. Value {value} was provided"
            )
        self._annealing_time = value

    # @property
    # def mixer_qubit_connectivity(self):
    #     return self._mixer_qubit_connectivity

    # @annealing_time.setter
    # def mixer_qubit_connectivity(self, value):
    #     print(value)
    #     if (self.mixer_hamiltonian != 'xy') and (value != None):
    #         self._mixer_qubit_connectivity = None
    #         raise ValueError(f"mixer_qubit_connectivity can be used if and only if `mixer_hamiltonian` is set to `xy`")
    #     else:
    #         print(value)
    #         self._mixer_qubit_connectivity = value


class BackendProperties(WorkflowProperties):
    """
    Choose the backend on which to run the QAOA circuits

    Parameters
    ----------
    device: DeviceBase
        The device to use for the backend.
    prepend_state: Union[openqaoa.basebackend.QuantumCircuitBase,numpy.ndarray(complex)]
        The state prepended to the circuit.
    append_state: Union[QuantumCircuitBase,numpy.ndarray(complex)]
        The state appended to the circuit.
    init_hadamard: bool
        Whether to apply a Hadamard gate to the beginning of the
        QAOA part of the circuit.
    n_shots: int
        The number of shots to be used for the shot-based computation.
    cvar_alpha: float
        The value of the CVaR parameter.
    noise_model: NoiseModel
        The `qiskit` noise model to be used for the shot-based simulator.
    initial_qubit_mapping: Union[List[int], numpy.ndarray]
        Mapping from physical to logical qubit indices, used to eventually
        construct the quantum circuit.  For example, for a system composed by 3 qubits
       `qubit_layout=[1,3,2]`, maps `1<->0`, `3<->1`, `2<->2`, where the left hand side is the physical qubit
        and the right hand side is the logical qubits
    qiskit_simulation_method: str
        Specify the simulation method to use with the `qiskit.AerSimulator`
    qiskit_optimization_level: int, optional
        Specify the qiskit.transpile optimization level. Choose from 0,1,2,3
    seed_simulator: int
        Specify a seed for `qiskit` simulators
    active_reset: bool
        To use the active_reset functionality on Rigetti backends through QCS
    rewiring: str
        Specify the rewiring strategy for compilation for Rigetti QPUs through QCS
    disable_qubit_rewiring: bool
        enable/disbale qubit rewiring when accessing QPUs via the AWS `braket`
    """

    def __init__(
        self,
        prepend_state: Optional[
            Union[QuantumCircuitBase, List[complex], np.ndarray]
        ] = None,
        append_state: Optional[Union[QuantumCircuitBase, np.ndarray]] = None,
        init_hadamard: bool = True,
        n_shots: int = 100,
        cvar_alpha: float = 1,
        noise_model=None,
        initial_qubit_mapping: Optional[Union[List[int], np.ndarray]] = None,
        qiskit_simulation_method: Optional[str] = None,
        qiskit_optimization_level: Optional[int] = None,
        seed_simulator: Optional[int] = None,
        active_reset: Optional[bool] = None,
        rewiring: Optional[str] = None,
        disable_qubit_rewiring: Optional[bool] = None,
    ):
        self.init_hadamard = init_hadamard
        self.n_shots = n_shots
        self.prepend_state = prepend_state
        self.append_state = append_state
        self.cvar_alpha = cvar_alpha
        self.noise_model = noise_model
        self.initial_qubit_mapping = initial_qubit_mapping
        self.seed_simulator = seed_simulator
        self.qiskit_simulation_method = qiskit_simulation_method
        self.qiskit_optimization_level = qiskit_optimization_level
        self.active_reset = active_reset
        self.rewiring = rewiring
        self.disable_qubit_rewiring = disable_qubit_rewiring

    # @property
    # def cvar_alpha(self):
    #     return self._cvar_alpha

    # @cvar_alpha.setter
    # def cvar_alpha(self, value):
    #     if (value <0) or (value>1) :
    #         raise ValueError(
    #             f"cvar_alpha must be between 0 and 1. Received {value}.")
    #     self._cvar_alpha = value


class ErrorMitigationProperties(WorkflowProperties):
    """
    Optional, choose an error mitigation technique for the QAOA circuit. Currently supports only SPAM Twirling.

    Parameters
    ----------
    error_mitigation_technique: str
        The name of the error mitigation technique. Only 'spam_twirling' supported for now.
    n_batches: Optional[int] = int
        Number of batches in which the total number of shots is divided to. For every batch, we choose a set of qubits at random to which we apply X gates and classical negating. The dafault value is set to 10 to be comparable with most problem sizes in NISQ without creating too much of an overhead.
    calibration_data_location: str
        The path to the file containing calibration data for the specific device.
    """

    def __init__(
        self,
        error_mitigation_technique: Optional[str] = None,
        n_batches: Optional[int] = 10,
        calibration_data_location: Optional[str] = None,
    ):
        self.error_mitigation_technique = (
            error_mitigation_technique.lower()
            if type(error_mitigation_technique) == str
            else error_mitigation_technique
        )

        if isinstance(n_batches, int) and n_batches > 0:
            self.n_batches = n_batches
        else:
            raise ValueError("n_batches must be a positive integer.")

        if calibration_data_location != None:
            try:
                with open(calibration_data_location, "r") as file:
                    # Parse the JSON file
                    calibration_data = json.load(file)

                    # Check if the file has the expected structure
                    calibration_measurements = calibration_data["results"][
                        "measurement_outcomes"
                    ]
                    calibration_registers = calibration_data["register"]

            except FileNotFoundError:
                raise FileNotFoundError(
                    "Calibration data file not found at specified location: {}".format(
                        calibration_data_location
                    )
                )
            except ValueError:
                raise ValueError(
                    "Calibration data file {} is not a valid JSON file".format(
                        calibration_data_location
                    )
                )
            except KeyError:
                raise KeyError(
                    "Calibration data file {} structure not as expected".format(
                        calibration_data_location
                    )
                )

        self.calibration_data_location = calibration_data_location


class ClassicalOptimizer(WorkflowProperties):
    """
    The classical optimizer for the QAOA optimization routine
    of the QAOA circuit parameters.

    Parameters
    ----------
    optimize: bool
        Whether to perform optimization routine on the given QAOA problem
    method: str
        optimization method for QAOA e.g. 'COBYLA'
    maxiter : Optional[int]
        Maximum number of iterations.
    maxfev : Optional[int]
        Maximum number of function evaluations.
    jac: str
        Method to compute the gradient vector. Choose from:
            - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
    hess:
        Method to compute the hessian. Choose from:
            - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
    constraints: `scipy.optimize.LinearConstraints`, `scipy.optimize.NonlinearConstraints`
        Scipy-based constraints on parameters of optimization
    bounds: `scipy.scipy.optimize.Bounds`
        Scipy-based bounds on parameters of optimization
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than
        the difference between two steps, terminate optimization.
    optimizer_options: dict
        Dictionary of optimiser-specific arguments, defaults to ``None``.
            stepsize : float
                Step size of each gradient descent step.
            decay : float
                Stepsize decay parameter of RMSProp.
            eps : float
                Small number to prevent division by zero for RMSProp.
            lambd : float
                Small number to regularize QFIM for Natural Gradient Descent.
    jac_options : dict
        Dictionary that specifies gradient-computation options according to method chosen in 'jac'.
    hess_options : dict
        Dictionary that specifies Hessian-computation options according to method chosen in 'hess'.
    optimization_progress : bool
        Returns history of measurement outcomes/wavefunction if `True`. Defaults to `False`.
    cost_progress : bool
        Returns history of cost values if `True`. Defaults to `True`.
    parameter_log : bool
        Returns history of angles if `True`. Defaults to `True`.
    save_intermediate: bool
        Outputs the jobids and parameters used for each circuit into
        seperate csv files. Defaults to `False`.
    """

    def __init__(
        self,
        optimize: bool = True,
        method: str = "cobyla",
        maxiter: int = 100,
        maxfev: int = None,
        jac: str = None,
        hess: str = None,
        constraints=None,
        bounds=None,
        tol=None,
        optimizer_options: dict = None,
        jac_options: dict = None,
        hess_options: dict = None,
        optimization_progress: bool = False,
        cost_progress: bool = True,
        parameter_log: bool = True,
        save_intermediate: bool = False,
    ):
        self.optimize = optimize
        self.method = method.lower()
        self.maxiter = maxiter
        self.maxfev = maxfev
        self.jac = jac.lower() if type(jac) == str else jac
        self.hess = hess.lower() if type(hess) == str else hess
        self.constraints = constraints
        self.bounds = bounds
        self.tol = tol
        self.optimizer_options = optimizer_options
        self.jac_options = jac_options
        self.hess_options = hess_options
        self.parameter_log = parameter_log
        self.optimization_progress = optimization_progress
        self.cost_progress = cost_progress
        self.parameter_log = parameter_log
        self.save_intermediate = save_intermediate

    # @property
    # def method(self):
    #     return self._method

    # @method.setter
    # def method(self, value):
    #     if value not in ALLOWED_MINIMIZATION_METHODS:
    #         raise ValueError(
    #             f"method `{value}` is not supported. Please choose between {ALLOWED_MINIMIZATION_METHODS}")
    #     self._method = value
